import pytest


@pytest.fixture
def subclass_registry_module():
    from software_patterns import SubclassRegistry
    from software_patterns.subclass_registry import InstantiationError, \
        UnknownClassError
    return type('M', (), {
        'SubclassRegistry': SubclassRegistry,
        'InstantiationError': InstantiationError,
        'UnknownClassError': UnknownClassError,
    })


@pytest.fixture
def register_class(subclass_registry_module):
    def _register_class(subclass_id: str, inherit=False):
        class ParentClass(metaclass=subclass_registry_module.SubclassRegistry):
            pass

        if inherit:
            @ParentClass.register_as_subclass(subclass_id)
            class Child(ParentClass):
                pass
        else:
            @ParentClass.register_as_subclass(subclass_id)
            class Child:
                pass

        child_instance = ParentClass.create(subclass_id)

        return {'class_registry': ParentClass, 'child': Child, 'child_instance': child_instance}
    return _register_class


@pytest.fixture
def use_metaclass(register_class, assert_correct_metaclass_behaviour):
    inherited_from_parent = {True: lambda classes: isinstance(classes['child_instance'], classes['class_registry']),
                False: lambda classes: not isinstance(classes['child_instance'], classes['class_registry'])}
    def _use_metaclass_in_scenario(subclass_id: str, inherit=False):
        classes = register_class(subclass_id, inherit=inherit)
        assert_correct_metaclass_behaviour(classes, subclass_id)
        assert inherited_from_parent[inherit]
        return classes['child_instance'], classes['child'], classes['class_registry']
    return _use_metaclass_in_scenario


@pytest.fixture
def assert_correct_metaclass_behaviour():
    def assert_metaclass_behaviour(classes, subclass_id):
        assert classes['class_registry'].subclasses[subclass_id] == classes['child']
        assert type(classes['child_instance']) == classes['child']
        assert isinstance(classes['child_instance'], classes['child'])
    return assert_metaclass_behaviour


def test_metaclass_usage(subclass_registry_module):
    class ParentClass(metaclass=subclass_registry_module.SubclassRegistry):
        pass

    assert type(ParentClass) == subclass_registry_module.SubclassRegistry
    assert hasattr(ParentClass, 'subclasses')
    assert hasattr(ParentClass, 'create')
    assert hasattr(ParentClass, 'register_as_subclass')
    assert ParentClass.subclasses == {}


def test_subclass_registry(use_metaclass, subclass_registry_module):
    child1_instance1, Child1, ParentClass = use_metaclass('child1', inherit=True)

    non_existent_identifier = 'child2'

    exception_message_regex = \
        f'Bad "{str(ParentClass.__name__)}" subclass request; requested subclass with identifier ' \
        f'{non_existent_identifier}, but known identifiers are ' \
        rf'\[{", ".join(subclass_identifier for subclass_identifier in ParentClass.subclasses.keys())}\]'

    with pytest.raises(subclass_registry_module.UnknownClassError, match=exception_message_regex):
        ParentClass.create(non_existent_identifier)

    child1_instance2, Child2, ParentClass2 = use_metaclass('child2', inherit=False)
    assert ParentClass.subclasses['child1'] == Child1


def test_create_wrong_input(subclass_registry_module):

    class ClassRegistry(metaclass=subclass_registry_module.SubclassRegistry):
            pass

    @ClassRegistry.register_as_subclass('id-1')
    class ConcreteClassA: pass

    @ClassRegistry.register_as_subclass('id-2')
    class ConcreteClassB:
        def __init__(self, a):
            self.a = a

    with pytest.raises(subclass_registry_module.InstantiationError):
        ClassRegistry.create('id-1', 'extra-argument')

    with pytest.raises(subclass_registry_module.InstantiationError):
        ClassRegistry.create('id-1', extra_key='extra-kwarg')

    ClassRegistry.create('id-2', 'argument-a-provided')
    with pytest.raises(subclass_registry_module.InstantiationError):
        ClassRegistry.create('id-2')



import pytest


@pytest.fixture
def simple_memoize():
    from software_patterns import ObjectsPool

    class TestClass:
        def __init__(self, number, name):
            self.number = number
            self.name = name

        @staticmethod
        def factory_method(arg1, arg2, **kwargs):
            return TestClass(arg1, arg2)
    return ObjectsPool(TestClass.factory_method)


def test_simple_memoize(simple_memoize):
    runtime_args = (7, 'gg')
    runtime_kwargs = dict(kwarg1='something', kwarg2=[1, 2])
    instance1 = simple_memoize.get_object(*runtime_args, **runtime_kwargs)
    instance2 = simple_memoize.get_object(*runtime_args, **runtime_kwargs)
    assert instance1 == instance2
    assert id(instance1) == id(instance2)
    hash1 = simple_memoize._build_hash(*runtime_args, **runtime_kwargs)
    assert hash1 == hash('-'.join([str(_) for _ in runtime_args] + [f'{key}={str(value)}' for key, value in runtime_kwargs.items()]))



"""Exposes the SubclassRegistry that allows to define a single registration point of one or more subclasses of a
(common parent) class.
"""

from typing import TypeVar, Generic, Dict

T = TypeVar('T')


class SubclassRegistry(type, Generic[T]):
    """Subclass Registry

    A (parent) class using this class as metaclass gains the 'subclasses' class attribute as well as the 'create' and
    'register_as_subclass' class methods.

    The 'subclasses' attribute is a python dictionary having string identifiers as keys and subclasses of the (parent)
    class as values.

    The 'register_as_subclass' class method can be used as a decorator to indicate that a (child) class should belong in
    the parent's class registry. An input string argument will be used as the unique key to register the subclass.

    The 'create' class method can be invoked with a (string) key and suitable constructor arguments to later construct
    instances of the corresponding child class.

    Example:

        >>> from software_patterns import SubclassRegistry

        >>> class ClassRegistry(metaclass=SubclassRegistry):
        ...  pass

        >>> ClassRegistry.subclasses
        {}

        >>> @ClassRegistry.register_as_subclass('child')
        ... class ChildClass:
        ...  def __init__(self, child_attribute):
        ...   self.attr = child_attribute

        >>> child_instance = ClassRegistry.create('child', 'attribute-value')
        >>> child_instance.attr
        'attribute-value'

        >>> type(child_instance).__name__
        'ChildClass'

        >>> isinstance(child_instance, ChildClass)
        True

        >>> {k: v.__name__ for k, v in ClassRegistry.subclasses.items()}
        {'child': 'ChildClass'}
    """
    subclasses: Dict[str, type]

    def __init__(cls, *args):
        super().__init__(*args)
        cls.subclasses = {}

    def create(cls, subclass_identifier, *args, **kwargs) -> T:
        """Create an instance of a registered subclass, given its unique identifier and runtime (constructor) arguments.

        Invokes the identified subclass constructor passing any supplied arguments. The user needs to know the arguments
        to supply depending on the resulting constructor signature.

        Args:
            subclass_identifier (str): the unique identifier under which to look for the corresponding subclass

        Raises:
            UnknownClassError: In case the given identifier is unknown to the parent class
            InstantiationError: In case the runtime args and kwargs do not match the constructor signature

        Returns:
            object: the instance of the registered subclass
        """
        if subclass_identifier not in cls.subclasses:
            raise UnknownClassError(f'Bad "{str(cls.__name__)}" subclass request; requested subclass with identifier '
                             f'{str(subclass_identifier)}, but known identifiers are '
                             f'[{", ".join(str(subclass_id) for subclass_id in cls.subclasses.keys())}]')
        try:
            return cls.subclasses[subclass_identifier](*args, **kwargs)
        except Exception as any_error:
            raise InstantiationError('Error during instance object construction.'
                f' Failed to create instance of class {cls.subclasses[subclass_identifier]}'
                f' using args [{", ".join((str(_) for _ in args))}]'
                f' and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]') from any_error

    def register_as_subclass(cls, subclass_identifier):
        """Register a class as subclass of the parent class.

        Adds the subclass' constructor in the registry (dict) under the given (str) identifier. Overrides the registry
        in case of "identifier collision". Can be used as a python decorator.

        Args:
            subclass_identifier (str): the user-defined identifier, under which to register the subclass
        """
        def wrapper(subclass):
            """Add the (sub) class provided to the parent class registry.

            Args:
                subclass ([type]): the (sub) class to register

            Returns:
                object: the (sub) class
            """
            cls.subclasses[subclass_identifier] = subclass
            return subclass
        return wrapper


class InstantiationError(Exception): pass
class UnknownClassError(Exception): pass



import pytest

@pytest.fixture
def proxy_module():
    from software_patterns import proxy
    return proxy


@pytest.fixture
def dummy_handle():
    def handle(self, *args, **kwargs):
        return f'{type(self).__name__} handle request with args [{", ".join((str(_) for _ in args))}] and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]'
    return handle


def test_proxy_behaviour(proxy_module, dummy_handle, capsys):
    prm = proxy_module

    # replicate client code that wants to use the proxy pattern

    # Derive from class RealSubject or use 'python overloading' (use a class 
    # that has a 'request' method with the same signature as ReadSubject.request)
    class ClientSubject(prm.ProxySubject):
        def request(self, *args, **kwargs):
            # delegate handling to the real real subject from client code
            # so frankly the request method here is also an adapter
            print(dummy_handle(self, *args, **kwargs))
            super().request(*args, **kwargs)
            return type(self).__name__

    # Derive from Proxy
    class ClientProxy(prm.Proxy):

        def request(self, *args, **kwargs):

            # run proxy code before sending request to the "proxied" handler
            before_args = list(['before'] + list(args))
            print(dummy_handle(self, *before_args, **kwargs))

            # handle request with the proxied logic
            _ = super().request(*args, **kwargs)
            assert _ == 'ClientSubject'

            # run proxy code after request to the "proxied" handler
            after_args = list(['after'] + list(args))
            print(dummy_handle(self, *after_args, **kwargs))
            return _
    def _dummy_callback(*args, **kwargs):
        return None
    real_subject = ClientSubject(_dummy_callback)
    proxy = ClientProxy(real_subject)

    # use proxy in a scenario

    # First test what happens without using proxy
    args = [1, 2]
    kwargs = {'k1': 'v1'}
    result = real_subject.request(*args, **kwargs)

    captured = capsys.readouterr()
    assert captured.out == dummy_handle(real_subject, 1, 2, k1='v1') + '\n'
    assert captured.out == f'ClientSubject handle request with args [{", ".join(str(_) for _ in args)}] and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]\n'
    assert result == type(real_subject).__name__
    assert result == 'ClientSubject'

    # Now test what happens using proxy
    result = proxy.request(*args, **kwargs)

    captured = capsys.readouterr()
    assert captured.out == dummy_handle(*list([proxy, 'before'] + args), **kwargs) +'\n' + \
        dummy_handle(*list([real_subject] + args), **kwargs) + '\n' + \
        dummy_handle(*list([proxy, 'after'] + args), **kwargs) +'\n'
    assert captured.out == f'ClientProxy handle request with args [{", ".join(str(_) for _ in ["before"] + args)}] and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]\n' + \
        f'ClientSubject handle request with args [{", ".join(str(_) for _ in args)}] and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]\n' + \
        f'ClientProxy handle request with args [{", ".join(str(_) for _ in ["after"] + args)}] and kwargs [{", ".join(f"{k}={v}" for k, v in kwargs.items())}]\n'
    assert result == type(real_subject).__name__
    assert result == 'ClientSubject'



"""Implementation of the Memoize Software Design Pattern.

Memoize is implemented using an Object Pool which is queried by a key which is
the result of computing a hash given runtime arguments.

"""
from typing import Dict, Generic, TypeVar, Any, Callable, Optional, Union
import types


__all__ = ['ObjectsPool']



DictKey = Union[int, str]
ObjectType = TypeVar('ObjectType')

RuntimeBuildHashCallable = Callable[..., Union[int, str]]


def adapt_build_hash(a_callable: RuntimeBuildHashCallable):
    def build_hash(_self: ObjectType, *args, **kwargs):
        return a_callable(*args, **kwargs)
    return build_hash


class ObjectsPool(Generic[ObjectType]):
    """Cache objects and allow to query (the pool) using runtime arguments.

    Instances of the ObjectsPool class implement the Object Pool Software Design
    Creational Pattern.

    Whenever an object is requested, it is checked whether it exists in the
    pool by using the runtimetime arguments to query a python dictionary.
    If it exists, a reference is returned, otherwise a new object is
    constructed (given the provided callback) and its reference is returned.

    Example:
        >>> from software_patterns import ObjectsPool
        >>> class ClientClass:
        ...  def __init__(self, a: int, b: int):
        ...   pass

        >>> object_pool = ObjectsPool[ClientClass](ClientClass)

        >>> obj1 = object_pool.get_object(1, 2)
        >>> obj2 = object_pool.get_object(1, 3)
        >>> obj3 = object_pool.get_object(1, 2)

        >>> id(obj1) == id(obj3)
        True

        >>> len(object_pool._objects)
        2

    Args:
        callback (Callable[..., ObjectType]): constructs objects given arguments
        hash_callback (Optional[RuntimeBuildHashCallable], optional): option to
            overide the default hash key computer. Defaults to None.
    Returns:
        [type]: [description]
    """
    _objects: Dict[DictKey, ObjectType]

    user_supplied_callback: Dict[bool, Callable] = {
        True: lambda callback: callback,
        False: lambda callback: ObjectsPool.__build_hash,
    }

    def __init__(self, callback: Callable[..., ObjectType], hash_callback: Optional[RuntimeBuildHashCallable]=None):
        self.constructor = callback
        build_hash_callback = self.user_supplied_callback[callable(hash_callback)](hash_callback)
        self._build_hash = types.MethodType(adapt_build_hash(build_hash_callback), self)
        self._objects = {}

    @staticmethod
    def __build_hash(*args: Any, **kwargs: Any) -> int:
        r"""Construct a hash out of the input \*args and \*\*kwargs."""
        return hash('-'.join([str(_) for _ in args] + [f'{key}={str(value)}' for key, value in kwargs.items()]))

    def get_object(self, *args: Any, **kwargs: Any) -> ObjectType:
        r"""Request an object from the pool.

        Get or create an object given the input arguments, which are used to
        create a unique hash key. The key is used to query a python dictionary
        and determine whether the object request refers to a cached object.

        Returns:
            object (ObjectType): the reference to the object that corresponds to the input
            arguments, regardless of whether it was found in the pool or not
        """
        key = self._build_hash(*args, **kwargs)
        if key not in self._objects:
            self._objects[key] = self.constructor(*args, **kwargs)
        return self._objects[key]



"""Notification-Listener (aka subject-observer) software design pattern.

Simple implementation of the subject/observers (broadcast/listeners) pattern,
exposed as python classes.

This is a Behavioural Pattern which can be used when you want one or more
components to be notified and react accordingly, when 'something happens'.

One entity, known as Subject, is responsible to send out a notification and each
entity "subscribed" to the Subject receives it and reacts.
Each subscribed entity is known as a Listener or Observer.

The idea is that the Subject is agnostic of its Observers implementation and the
client code can "attach" or "detach" (subscribe/unsubscribe) as many of them at
runtime.

This module provides the Observer class, serving as the interface that needs to
be implemented by concrete classes; the update method needs to be overrode.
Concrete Observers react to the notifications/updates issued by the Subject they
had been attached/subscribed to.

This module also provides the concrete Subject class, serving with methods to
subscribe/unsubscribe (attach/detach) observers and also with a method to
"notify" all Observers.
"""

from abc import ABC, abstractmethod
from typing import List, Generic, TypeVar

__all__ = ['Subject', 'Observer']


StateType = TypeVar('StateType')


class ObserverInterface(ABC):
    """The Observer interface declares the update method, used by subjects.

    Enables objects to act as "event" listeners; react to "notifications"
    by executing specific (event) handling logic.
    """
    @abstractmethod
    def update(self, *args, **kwargs) -> None:
        """Receive an update (from a subject); handle an event notification."""
        raise NotImplementedError


class SubjectInterface(ABC):
    """The Subject interface declares a set of methods for managing subscribers.

    Enables objects to act as subjects that broadcast events" by notifying
    all subscribed observers/listeners.
    """

    @abstractmethod
    def attach(self, observer: ObserverInterface) -> None:
        """Attach an observer to the subject; subscribe the observer."""
        raise NotImplementedError

    @abstractmethod
    def detach(self, observer: ObserverInterface) -> None:
        """Detach an observer from the subject; unsubscribe the observer."""
        raise NotImplementedError

    @abstractmethod
    def notify(self) -> None:
        """Notify all observers about an event."""
        raise NotImplementedError


class Observer(ObserverInterface, ABC):
    pass


class Subject(SubjectInterface, Generic[StateType]):
    """Concrete Subject which owns an important state and notifies observers.

    The subject can be used to build the data encapsulating the event being
    broadcasted.

    Both the _state and _observers attributes have a simple implementation,
    but can be overrode to accommodate for more complex scenarios.

    The observers/subscribers are implemented as a python list.
    In more complex scenarios, the list of subscribers can
    be stored more comprehensively (categorized by event type, etc.).

    The subscription management methods provided are 'attach', 'detach' (as in
    the SubjectInterface) and 'add', which attached multiple observers at once.

    Example:

        >>> from software_patterns import Subject, Observer

        >>> broadcaster = Subject()

        >>> class ObserverTypeA(Observer):
        ...  def update(self, *args, **kwargs):
        ...   event = args[0].state
        ...   print(f'observer-type-a reacts to event {event}')

        >>> class ObserverTypeB(Observer):
        ...  def update(self, *args, **kwargs):
        ...   event = args[0].state
        ...   print(f'observer-type-b reacts to event {event}')

        >>> subscriber_1 = ObserverTypeA()
        >>> subscriber_2 = ObserverTypeB()

        >>> broadcaster.add(subscriber_2, subscriber_1)

        >>> broadcaster.state = 'event-object-A'

        >>> broadcaster.notify()
        observer-type-b reacts to event event-object-A
        observer-type-a reacts to event event-object-A

        >>> broadcaster.detach(subscriber_2)

        >>> broadcaster.state = 'event-object-B'
        >>> broadcaster.notify()
        observer-type-a reacts to event event-object-B
    """
    def __init__(self, *args, **kwargs):
        self._observers: List[ObserverInterface] = []
        self._state = None

    def attach(self, observer: ObserverInterface) -> None:
        self._observers.append(observer)

    def detach(self, observer: ObserverInterface) -> None:
        self._observers.remove(observer)

    def notify(self) -> None:
        for observer in self._observers:
            observer.update(self)

    def add(self, *observers):
        """Subscribe multiple observers at once."""
        self._observers.extend(list(observers))

    @property
    def state(self) -> StateType:
        """Get the state of the Subject.

        Returns:
            StateType: the object representing the current state of the Subject
        """
        return self._state

    @state.setter
    def state(self, state: StateType):
        """Set the state of the Subject.

        Args:
            state (StateType): the state object
        """
        self._state = state



import pytest


@pytest.fixture
def subject():
    from software_patterns import Subject
    return Subject


@pytest.fixture
def observer():
    from software_patterns import Observer
    return Observer


def test_observers_sanity_test1(subject):
    subject1 = subject([])
    subject2 = subject([])
    assert hasattr(subject1, '_observers')
    assert hasattr(subject2, '_observers')
    assert id(subject1._observers) != id(subject2._observers)


def test_observer_as_constructor(observer):
    with pytest.raises(TypeError) as instantiation_from_interface_error:
        _observer_instance = observer()

    import re
    runtime_exception_message_reg = "Can't instantiate abstract class " \
        "Observer with abstract methods? update"

    assert re.match(
        runtime_exception_message_reg,
        str(instantiation_from_interface_error.value)
    )

def test_scenario(subject, observer):
# The client code.

    print("------ Scenario 1 ------\n")
    class ObserverA(observer):
        def update(self, a_subject) -> None:
            print("ObserverA: Reacted to the event")

    s1 = subject([])
    o1 = ObserverA()
    s1.attach(o1)

    # business logic
    s1.state = 0
    s1.notify()

    print("------ Scenario 2 ------\n")
    # example 2
    class Businessubject(subject):

        def some_business_logic(self) -> None:
            """
            Usually, the subscription logic is only a fraction of what a Subject can
            really do. Subjects commonly hold some important business logic, that
            triggers a notification method whenever something important is about to
            happen (or after it).
            """
            print("\nSubject: I'm doing something important.")
            self._state = 2
            print(f"Subject: My state has just changed to: {self._state}")
            self.notify()

    class ObserverB(observer):
        def update(self, a_subject) -> None:
            if a_subject.state == 0 or a_subject.state >= 2:
                print("ObserverB: Reacted to the event")

    s2 = Businessubject([])
    assert id(s1) != id(s2)
    assert id(s1._observers) != id(s2._observers)
    o1, o2 = ObserverA(), ObserverB()

    s2.add(o1, o2)
    # business logic
    print(s2._observers)
    s2.some_business_logic()
    s2.some_business_logic()

    s2.detach(o1)
    s2.some_business_logic()



__version__ = '1.1.3'

from .notification import Observer, Subject
from .memoize import ObjectsPool
from .proxy import ProxySubject, Proxy
from .subclass_registry import SubclassRegistry


__all__ = ['Observer', 'Subject', 'ObjectsPool', 'SubclassRegistry',
    'ProxySubject', 'Proxy']



"""Proxy structural software pattern.

This module contains boiler-plate code to supply the Proxy structural software
design pattern, to the client code."""

from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Callable

T = TypeVar('T')


__all__ = ['ProxySubject', 'Proxy']


class ProxySubjectInterface(ABC, Generic[T]):
    """Proxy Subject interface holding the important 'request' operation.

    Declares common operations for both ProxySubject and
    the Proxy. As long as the client uses ProxySubject's interface, a proxy can
    be passed pass to it, instead of a real subject.
    """
    @abstractmethod
    def request(self, *args, **kwargs) -> T:
        raise NotImplementedError


class ProxySubject(ProxySubjectInterface, Generic[T]):
    """
    The ProxySubject contains some core business logic. Usually, ProxySubject are
    capable of doing some useful work which may also be very slow or sensitive -
    e.g. correcting input data. A Proxy can solve these issues without any
    changes to the ProxySubject's code.

    Example:

        >>> from software_patterns import ProxySubject
        >>> proxied_operation = lambda x: x + 1
        >>> proxied_operation(1)
        2

        >>> proxied_object = ProxySubject(proxied_operation)
        >>> proxied_object.request(1)
        2
    """
    def __init__(self, callback: Callable[..., T]):
        self._callback = callback

    def request(self, *args, **kwargs) -> T:
        return self._callback(*args, **kwargs)


class Proxy(ProxySubjectInterface, Generic[T]):
    """
    The Proxy has an interface identical to the ProxySubject.

    Example:

        >>> from software_patterns import Proxy
        >>> from software_patterns import ProxySubject

        >>> class ClientProxy(Proxy):
        ...  def request(self, *args, **kwargs):
        ...   args = [args[0] + 1]
        ...   result = super().request(*args, **kwargs)
        ...   result += 1
        ...   return result

        >>> proxied_operation = lambda x: x * 2
        >>> proxy_subject = ProxySubject(proxied_operation)
        >>> proxy_subject.request(3)
        6

        >>> proxy = ClientProxy(proxy_subject)
        >>> proxy.request(3)
        9
    """
    def __init__(self, proxy_subject: ProxySubject):
        self._proxy_subject = proxy_subject

    def request(self, *args, **kwargs) -> T:
        """
        The most common applications of the Proxy pattern are lazy loading,
        caching, controlling the access, logging, etc. A Proxy can perform one
        of these things and then, depending on the result, pass the execution to
        the same method in a linked ProxySubject object.
        """
        return self._proxy_subject.request(*args, **kwargs)
